/* 
 * ==========================================================
 * Artekit test sketch for the AK-SST25VF080B 8 Mbit EEPROM
 * Copyright (c) 2013 Artekit Italy
 * http://www.artekit.eu
 *
 * Connect the pins in the following way:
 *
 * Arduino UNO          AK-SST25VF080B
 * =====================================
 *   Digital 13             CLK
 *   Digital 12             MISO
 *   Digital 11             MOSI
 *   Digital 10             nCS
 *   3.3V                   VCC
 *   GND                    GND
 *
 * This is a simple test sketch to evaluate the AK-SST25VF080B
 * breakout board. What this program does is to write 32 bytes
 * of randomly generated data into the first 32 bytes of each
 * sector. After that, the data is read and compared with the
 * original 32 bytes.
 * 
 * You may have to use a logic level converter, because the 
 * AK-SST25VF080B is not 5V tolerant.
 *
 * ==========================================================
 */
 
#include <SPI.h>

/* Buffer containing the random data */
byte write_buffer[32];

/* Buffer used to read */
byte read_buffer[32];

/* Current page (reading or writing) */
unsigned long page = 0;

/* This function reads the SST25VF080B status register */
byte ReadStatusRegister()
{
  byte val;
  digitalWrite(10, LOW);
  SPI.transfer(0x05);
  val = SPI.transfer(0x00);
  digitalWrite(10, HIGH);
  
  return val;
}

/* This function returns true if the SST25VF080B is busy
 * writing. Otherwise it returns false.
 */
bool IsBusy()
{
  if (ReadStatusRegister() & 0x01)
  {
    return true;
  }
  return false;
}

/* Enables write/erase operations */
void WriteEnable()
{
  digitalWrite(10, LOW);
  SPI.transfer(0x06);
  digitalWrite(10, HIGH);
}

/* By default, the SST25VF080B starts with all the sectors
 * write-protected. We disable the protection calling this
 * function */
void Unprotect()
{
  WriteEnable();

  digitalWrite(10, LOW);
  SPI.transfer(0x01);
  SPI.transfer(0x00);
  digitalWrite(10, HIGH);
}

/* Disables write/erase operations */
void WriteDisable()
{
  digitalWrite(10, LOW);
  SPI.transfer(0x04);
  digitalWrite(10, HIGH);
}

/* Erases a 4KBytes page (sets all 4KBytes to 0xFF) */
void Erase4KPage(unsigned long page)
{
  unsigned long addr = (unsigned long) page * 0x1000;
  WriteEnable();

  digitalWrite(10, LOW);
  SPI.transfer(0x20);
  SPI.transfer((addr >> 16) & 0xFF);
  SPI.transfer((addr >> 8) & 0xFF);
  SPI.transfer(addr & 0xFF);
  digitalWrite(10, HIGH);
  
  while (IsBusy());
}

/* Reads the given quantity of bytes from 'addr' address into
 * 'buf' buffer.
 * Returns the quantity of read bytes.
 */
unsigned long Read(unsigned long addr, byte* buf, unsigned long qty)
{
  unsigned long read = 0;
  
  while (IsBusy());
  
  digitalWrite(10, LOW);
  SPI.transfer(0x03);
  SPI.transfer((addr >> 16) & 0xFF);
  SPI.transfer((addr >> 8) & 0xFF);
  SPI.transfer(addr & 0xFF);
  
  while (read < qty)
  {
    buf[read++] = SPI.transfer(0x00);
  }
  
  digitalWrite(10, HIGH);
  return read;
}

/* Writes the 'qty' quantity of bytes from 'buf' buffer into
 * 'addr' address.
 * This function will use the AAI (Auto Address Increment) 
 * feature if qty >= 2.
 */
unsigned long Write(unsigned long addr, byte* buf, unsigned long qty)
{
  unsigned long cnt = 0;
  bool aai = false;
  
  /* Enable write operation */
  WriteEnable();
  
  while (cnt < qty)
  {
    /* Check if busy and wait for every write */
    while (IsBusy());
    
    if (qty - cnt >= 2)
    {
      /* Auto Address Increment */
      digitalWrite(10, LOW);

      SPI.transfer(0xAD);
      
      if (aai == false)
      {
        /* First time issuing AAI command; send
         * the destination address. */
        SPI.transfer((addr >> 16) & 0xFF);
        SPI.transfer((addr >> 8) & 0xFF);
        SPI.transfer(addr & 0xFF);
        aai = true;
      }

      /* If aai = true, then we don't need to send
       * the address again. It will be automatically
       * incremented.
       */
      SPI.transfer(buf[cnt++]);
      SPI.transfer(buf[cnt++]);

      digitalWrite(10, HIGH);

      /* Keep track of the last written address */
      addr += 2;
    } else {
      
      if (aai)
      {
        /* If an AAI was issued, we need to terminate
         * it by calling WriteDisable().
         */
        WriteDisable();
        /* And... since we're writing again, call.. */
        WriteEnable();
        aai = false;
      }
      
      /* Single-byte write */
      digitalWrite(10, LOW);
      
      SPI.transfer(0x02);
      SPI.transfer((addr >> 16) & 0xFF);
      SPI.transfer((addr >> 8) & 0xFF);
      SPI.transfer(addr & 0xFF);
      SPI.transfer(buf[cnt++]);
     
      digitalWrite(10, HIGH);
    }       
  }
  
  /* Disable write operations */
  WriteDisable();
  return cnt;
}

void setup()
{
  /* Start serial */ 
  Serial.begin(57600);
  /* Start and configure SPI */
  SPI.begin();
  SPI.setBitOrder(MSBFIRST);
  /* The datasheet says that SPI_MODE3 may also work */
  SPI.setDataMode(SPI_MODE0);
  
  Unprotect();
}

void loop()
{
  /* Seed the random number generator with noise from
   * a disconnected AD input */
  randomSeed(analogRead(0));
  
  /* Generate 32 random numbers */
  Serial.println("32 bytes pattern:");  
  for (int i = 0; i < 32; i++)
  {  
    write_buffer[i] = random(0xFF); 
    Serial.print(write_buffer[i]);
    Serial.print(" ");
  }

  Serial.println(" ");
  
  /* Write pages */
  while (page < 244)
  {
      Erase4KPage(page);
      Write(page * 0x1000, write_buffer, 32);
      Serial.print("Writing page ");
      Serial.println(page);
      page++;
  }
  
  /* Read pages */
  page = 0;
  while (page < 244)
  {
    Serial.println(" ");
    Serial.print("Reading page ");
    Serial.println(page);
    
    memset(read_buffer, 0, 32);
    Read(page * 0x1000, read_buffer, 32);
    
    for (int i = 0; i < 32; i++)
    {
       if ((i+1) % 16 == 0)
       {
         Serial.println(read_buffer[i]);
  
       } else {     
         Serial.print(read_buffer[i]); 
         Serial.print(" ");       
       }
    }
    
    /* Compare the read buffer with the written buffer */
    if (memcmp(read_buffer, write_buffer, 32) == 0)
    {
      /* Matches, everything is OK */
      Serial.println("Page matches");
    } else {
      /* Read/Write error */
      Serial.print("Read/Write error on page ");
      Serial.println(page);
      while(1);
    }
    
    page++;
  }

  /* Stop */
  while(1);
}

